using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using MeshVR;
using MVR.FileManagementSecure;


public class FaceBlend : MVRScript
{
    private const string ASSET_BUNDLE_LOCATION = "Custom/Assets/northernshikima/northernshikima_faceblend.assetbundle";
    private const string DEFAULT_MASK_PATH = "Custom/Atom/Person/Textures/FaceBlend Masks";
    private const string DEFAULT_MASK_TEXTURE = "Female Face Blend Mask.png";

    private DAZCharacterSelector selector;
    private DAZCharacter dazCharacter;

    public static readonly List<string> TextureSlots = new List<string>() { "Torso", "Genitals", "Face", "Limbs" };

    private Shader unlitBlendShader;


    private static class BodyRegionEnum
    {
        public const string Torso = "Torso";
        public const string Face = "Face";
        public const string Limbs = "Limbs";
        public const string Genitals = "Genitals";
        public static readonly List<string> Values = new List<string>() { Torso, Genitals, Face, Limbs };
    }


    /*  TODO: Provide an apply textures on bake option
       private static readonly Dictionary<string, string> bodyRegionUrlParams = new Dictionary<string, string>()
        {
            {BodyRegionEnum.Torso, "torsoDiffuseUrl" },
            {BodyRegionEnum.Face, "faceDiffuseUrl" },
            {BodyRegionEnum.Limbs, "limbsDiffuseUrl" },
            {BodyRegionEnum.Genitals, "genitalsDiffuseUrl" },
        };*/

    private static readonly Dictionary<string, int[]> femaleTextureIndex = new Dictionary<string, int[]>()
        {
            {BodyRegionEnum.Torso, new int[]{15,18,19,20,21,27,29}},
            {BodyRegionEnum.Face, new int[]{2,5,11} },
            {BodyRegionEnum.Limbs, new int[]{0,12,14,16,17,22,23}},
            {BodyRegionEnum.Genitals, new int[]{28}},
        };

    private static readonly Dictionary<string, int[]> maleTextureIndex = new Dictionary<string, int[]>()
        {
            {BodyRegionEnum.Torso, new int[]{15,16,19,20,21,22,30,32}},
            {BodyRegionEnum.Face, new int[]{2,5,11}},
            {BodyRegionEnum.Limbs, new int[]{0,12,14,17,18,23,24}},
            {BodyRegionEnum.Genitals, new int[]{28,29}},
        };


    private Dictionary<string, Texture2D> originalTextures = new Dictionary<string, Texture2D>();
    private Dictionary<string, RenderTexture> renderTextures = new Dictionary<string, RenderTexture>();
    private Texture2D secondTexture;
    private Texture2D maskTexture = Texture2D.blackTexture;
    private Coroutine setupCoroutine;

    private bool setupComplete = false;
    private bool destroyedFromBake = false;


    // JSON Storables
    private JSONStorableStringChooser materialSlot = new JSONStorableStringChooser("Material Slot", BodyRegionEnum.Values, BodyRegionEnum.Face, "Material Slot");
    private JSONStorableUrl secondTextureURL = new JSONStorableUrl("FaceBlend Second Texture", string.Empty);
    private JSONStorableUrl maskTextureURL = new JSONStorableUrl("FaceBlend Mask Texture", string.Empty);
    private JSONStorableFloat hue = new JSONStorableFloat("Hue", 0.5f, 0.0f, 1.0f);
    private JSONStorableFloat saturation = new JSONStorableFloat("Saturation", 0.5f, 0.0f, 1.0f);
    private JSONStorableFloat value = new JSONStorableFloat("Value", 0.5f, 0.0f, 1.0f);

    private JSONStorableBool useAlphaAsMask = new JSONStorableBool("Use Alpha As Mask", false);
    private JSONStorableBool invertMask = new JSONStorableBool("Invert Mask", false);
    private JSONStorableBool adjustBase = new JSONStorableBool("Apply HSV to Base Textures", false);

    public override void Init()
    {
        try
        {
            if (containingAtom == null || containingAtom.type != "Person")
            {
                SuperController.LogError("This only works with person Atoms");
                DestroyImmediate(this);
                return;
            }

            selector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
            dazCharacter = selector.selectedCharacter;

            materialSlot.setCallbackFunction += UpdateMaterialSlot;
            RegisterStringChooser(materialSlot);
            CreatePopup(materialSlot);

            RegisterUrl(secondTextureURL);
            secondTextureURL.setCallbackFunction += LoadSecondTexture;

            RegisterUrl(maskTextureURL);
            maskTextureURL.setCallbackFunction += LoadMaskTexture;


            if (maskTextureURL.val == string.Empty)
            {
                maskTextureURL.SetFilePath(GetPackagePath(this) + DEFAULT_MASK_PATH + "/" + DEFAULT_MASK_TEXTURE);
                LoadMaskTexture(GetPackagePath(this) + DEFAULT_MASK_PATH + "/" + DEFAULT_MASK_TEXTURE);

            }

            UIDynamicButton buttonSecondTex = CreateButton("Browse Overlay Texture", false);
            buttonSecondTex.button.onClick.AddListener(() =>
            {
                // Todo: Most recent DIR, or default
                List<ShortCut> shortCuts = FileManagerSecure.GetShortCutsForDirectory("Custom/Atom/Person/Textures", false, true, true, false);
                SuperController.singleton.GetMediaPathDialog((string path) => { secondTextureURL.val = path; }, "jpg|png|tif|tiff", "Custom/Atom/Person/Textures", false, true, false, null, true, shortCuts, true, true);

            });
            UIDynamicTextField secondTextureLabel = CreateTextField(secondTextureURL, false);

            UIDynamicButton buttonMaskTex = CreateButton("Browse Mask Texture", false);
            buttonMaskTex.button.onClick.AddListener(() =>
            {
                // Todo: Most recent DIR, or default
                string defaultPath = GetPackagePath(this) + DEFAULT_MASK_PATH;
                List<ShortCut> shortCuts = FileManagerSecure.GetShortCutsForDirectory("Custom/Atom/Person/Textures", false, true, true, false);
                SuperController.singleton.GetMediaPathDialog((string path) => { maskTextureURL.val = path; }, "jpg|png|tif|tiff", defaultPath, false, true, false, null, true, shortCuts, true, true);

            });
            UIDynamicTextField maskTextureLabel = CreateTextField(maskTextureURL, false);


            // Setup HSL variables
            RegisterFloat(hue);
            hue.setCallbackFunction += (float val) => UpdateRenderTextures();
            CreateSlider(hue);
            RegisterFloat(saturation);
            saturation.setCallbackFunction += (float val) => UpdateRenderTextures();
            CreateSlider(saturation);
            RegisterFloat(value);
            value.setCallbackFunction += (float val) => UpdateRenderTextures();
            CreateSlider(value);

            UIDynamicButton bakeTextureButton = CreateButton("Bake Texture", false);
            bakeTextureButton.button.onClick.AddListener(() =>
            {
                SuperController.singleton.fileBrowserUI.defaultPath = "Custom/Atom/Person/Textures";
                SuperController.singleton.fileBrowserUI.showFiles = true;
                SuperController.singleton.fileBrowserUI.showDirs = true;
                SuperController.singleton.fileBrowserUI.SetTextEntry(true);
                SuperController.singleton.fileBrowserUI.hideExtension = true;
                SuperController.singleton.fileBrowserUI.shortCuts = new List<ShortCut>();
                SuperController.singleton.fileBrowserUI.fileFormat = "png";
                SuperController.singleton.fileBrowserUI.Show(BakeSelectedSlot);
            });

            UIDynamicButton bakeAllTexturesButton = CreateButton("Bake All Textures", false);
            bakeAllTexturesButton.button.onClick.AddListener(() =>
            {
                SuperController.singleton.fileBrowserUI.defaultPath = "Custom/Atom/Person/Textures";
                SuperController.singleton.fileBrowserUI.showFiles = true;
                SuperController.singleton.fileBrowserUI.showDirs = true;
                SuperController.singleton.fileBrowserUI.SetTextEntry(true);
                SuperController.singleton.fileBrowserUI.hideExtension = true;
                SuperController.singleton.fileBrowserUI.shortCuts = new List<ShortCut>();
                SuperController.singleton.fileBrowserUI.fileFormat = "png";
                SuperController.singleton.fileBrowserUI.Show(BakeAllSlots);
            });




            RegisterBool(useAlphaAsMask);
            useAlphaAsMask.setCallbackFunction += (bool val) => UpdateRenderTextures();
            CreateToggle(useAlphaAsMask, true);

            RegisterBool(invertMask);
            invertMask.setCallbackFunction += (bool val) => UpdateRenderTextures();
            CreateToggle(invertMask, true);


            RegisterBool(adjustBase);
            adjustBase.setCallbackFunction += (bool val) => UpdateRenderTextures();
            CreateToggle(adjustBase, true);


            AssetLoader.AssetBundleFromFileRequest req = new AssetLoader.AssetBundleFromFileRequest()
            {
                path = GetPackagePath(this) + ASSET_BUNDLE_LOCATION,
                callback = LoadAssetBundle,
            };
            AssetLoader.QueueLoadAssetBundleFromFile(req);



            setupCoroutine = StartCoroutine(Setup());
        }
        catch (Exception e)
        {
            SuperController.LogError("Exception caught: " + e);
        }
    }



    private void UpdateMaterialSlot(string val)
    {
        if (secondTextureURL.val == string.Empty)
        {
            secondTexture = null;
        }
        UpdateRenderTextures();
    }

    private void LoadSecondTexture(string path)
    {
        secondTextureURL.SetFilePath(path);
        ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
        queuedImage.imgPath = path;
        queuedImage.forceReload = true;
        queuedImage.skipCache = true;
        queuedImage.compress = false;
        queuedImage.createMipMaps = false;
        queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
        {
            secondTexture = qi.tex;
            UpdateRenderTextures();
        };
        ImageLoaderThreaded.singleton.QueueImage(queuedImage);

    }

    private void LoadMaskTexture(string path)
    {
        maskTextureURL.SetFilePath(path);
        ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
        queuedImage.imgPath = path;
        queuedImage.forceReload = true;
        queuedImage.skipCache = true;
        queuedImage.compress = false;
        queuedImage.createMipMaps = false;
        queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
        {
            maskTexture = qi.tex;
            UpdateRenderTextures();
        };
        ImageLoaderThreaded.singleton.QueueImage(queuedImage);
    }

    private void LoadAssetBundle(AssetLoader.AssetBundleFromFileRequest request)
    {

        if (request.assetBundle == null)
        {
            SuperController.LogError("Unable to load Asset Bundle - northernshikima_faceblend.assetbundle");
            return;
        }
        unlitBlendShader = request.assetBundle.LoadAsset<Shader>("assets/shaders/UnlitBlendTextures.shader");

        // If Assets successfully loaded
        if (unlitBlendShader)
        {
            UpdateRenderTextures();
        }
        else
        {
            SuperController.LogError("UnlitBlendTextures Shader not found in asset bundle!");
        }
    }

    private IEnumerator Setup()
    {
        while (!dazCharacter.ready)
        {
            yield return null; // Wait for a frame
        }

        // Setup render textures
        foreach (string region in BodyRegionEnum.Values)
        {
            int[] matSlots = GetMaterialSlots(region);
            originalTextures[region] = (Texture2D)dazCharacter.skin.GPUmaterials[matSlots[0]].GetTexture("_MainTex");

            renderTextures[region] = new RenderTexture(originalTextures[region].height, originalTextures[region].width, 0, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB);
            renderTextures[region].enableRandomWrite = true;
            renderTextures[region].Create();

            foreach (int i in matSlots)
            {
                dazCharacter.skin.GPUmaterials[i].SetTexture("_MainTex", renderTextures[region]);
            }

        }

        // Setup state here
        setupComplete = true;
        UpdateRenderTextures();
    }


    private void UpdateRenderTextures()
    {
        if (!setupComplete)
            return;

        if (!secondTexture)
        {
            secondTexture = originalTextures[materialSlot.val];
        }


        foreach (string slot in BodyRegionEnum.Values)
        {
            Material blendMaterial = new Material(unlitBlendShader);

            blendMaterial.SetTexture("_BaseTex", originalTextures[slot]);
            blendMaterial.SetTexture("_SecondTex", materialSlot.val == slot ? secondTexture : null);
            blendMaterial.SetTexture("_MaskTex", materialSlot.val == slot ? maskTexture : Texture2D.blackTexture);
            blendMaterial.SetFloat("_Hue", hue.val + 0.5f);
            blendMaterial.SetFloat("_Saturation", saturation.val * 2);
            blendMaterial.SetFloat("_Value", value.val * 2);
            blendMaterial.SetInt("_UseAlphaAsMask", materialSlot.val == slot && useAlphaAsMask.val ? 1 : 0);
            blendMaterial.SetInt("_InvertMask", materialSlot.val == slot && invertMask.val ? 1 : 0);
            blendMaterial.SetInt("_ApplyHSVToBase", adjustBase.val ? 1 : 0);

            Graphics.Blit(null, renderTextures[slot], blendMaterial);
        }
    }

    private void BakeSelectedSlot(string path)
    {
        UpdateRenderTextures();

        SaveRenderTexture(path, materialSlot.val);
    }

    private void BakeAllSlots(string path)
    {
        UpdateRenderTextures();

        foreach (string slot in BodyRegionEnum.Values)
        {
            SaveRenderTexture(path + "_" + slot.ToLower(), slot);
        }


    }

    private void SaveRenderTexture(string path, string slot)
    {
        path += ".png";

        Texture2D tex = new Texture2D(renderTextures[slot].height, renderTextures[slot].width, TextureFormat.RGB24, false);
        RenderTexture previous = RenderTexture.active;
        RenderTexture.active = renderTextures[slot];
        tex.ReadPixels(new Rect(0, 0, renderTextures[slot].width, renderTextures[slot].height), 0, 0);
        tex.Apply();
        RenderTexture.active = previous;
        FileManagerSecure.WriteAllBytes(path, tex.EncodeToPNG());

        /* TODO: Provide an apply textures on bake option
             DAZCharacterTextureControl textureControl = dazCharacter.GetComponentInChildren<DAZCharacterTextureControl>();
             textureControl.SetUrlParamValue(bodyRegionUrlParams[slot], path);
        */


    }


    public void OnDisable()
    {

        if (!setupComplete)
            return;

        foreach (string materialSlot in BodyRegionEnum.Values)
        {
            if (!destroyedFromBake)
            {
                ResetOriginalTextures(materialSlot);
            }
            renderTextures[materialSlot].Release();
        }


        originalTextures.Clear();
        renderTextures.Clear();

        // Clean up here
        setupComplete = false;
    }

    private void ResetOriginalTextures(string materialSlot)
    {
        Texture tex = originalTextures[materialSlot];
        foreach (int id in GetMaterialSlots(materialSlot))
        {
            dazCharacter.skin.GPUmaterials[id].SetTexture("_MainTex", tex);
        }
    }

    public void OnEnable()
    {
        if (setupCoroutine != null)
        {
            StopCoroutine(setupCoroutine);
        }
        setupCoroutine = StartCoroutine(Setup());
    }

    public void OnDestroy()
    {
        OnDisable();
        StopAllCoroutines();
    }

    private int[] GetMaterialSlots(string slot)
    {
        if (dazCharacter.isMale)
        {
            return maleTextureIndex[slot];
        }
        else
        {
            return femaleTextureIndex[slot];
        }
    }

    // Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
    // Credit to MacGruber Utils shared under CC BY
    private static string GetPluginPath(MVRScript self)
    {
        string id = self.name.Substring(0, self.name.IndexOf('_'));
        string filename = self.manager.GetJSON()["plugins"][id].Value;
        return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
    }

    // Get path prefix of the package that contains our plugin.
    // Credit to MacGruber Utils shared under CC BY
    private static string GetPackagePath(MVRScript self)
    {
        string filename = GetPluginPath(self);
        int idx = filename.IndexOf(":/");
        if (idx >= 0)
            return filename.Substring(0, idx + 2);
        else
            return string.Empty;
    }
}